/*****************************************************************************/
/* Wii Nunchuck Interface implementation                                     */
/* Author: Joao Geada							Version: 1.0	March 2009   */
/* Copyright (c) 2009, Joao Geada           See end of file for terms of use */
/*****************************************************************************/

#include <propeller.h>
#include "WiiNunchuck.h"

// address of nunchuck (shifted into the msb 7 bits. Real i2c address is 0x52)
#define NUNCHUCK_ADDR 0xA4

// speed to run i2c scl clock. My nunchuck works reliably at this speed
#define I2C_FREQ 55000

static unsigned short i2c_delay;         // delay between scl edges @ I2C_FREQ

static unsigned int nunchuck_cmd_delay;  // used to calc delay between cmds to Nunchuck

static unsigned int scl_pin,
		     		scl_pinb,  // inverse bitmask for scl_pin
			 		sda_pin, 
			 		sda_pinb,  // inverse bitmask for sda_pin 
			 		i2c_pins;

// NOTE: have to include a special version of i2c protocol driver
//       because Wii Nunchuck seems to require a slower & more regular scl clock
//       than is provided by the existing i2c drivers. This version
//       is loosely based on the James Burrow's SPIN i2c driver, v1.3 

/** start driving the i2c bus */			 
static void i2cStart()
{
   // SDA goes high to low while SCL high
   DIRA &= ~i2c_pins; // i2c pins as inputs
   DIRA |= sda_pin;   // sda output
   OUTA &= sda_pinb;  // sda low
   while (!(INA & scl_pin)) {} // wait for SCL high
   wait(i2c_delay);
}

/** release the i2c bus */
static void i2cStop()
{
   // SDA goes low to high while SCL high
   OUTA |= scl_pin;
   OUTA |= sda_pin;
   DIRA &= ~i2c_pins; // i2c pins as inputs -> pullups make both high
   wait(i2c_delay);
}

/** write 1 byte to the i2c bus */
static char i2cWrite(char dat)
{
   int ack = 0;
   int i;
   const unsigned int delay = i2c_delay; // local copy of i2c_delay for use in WAITCNT
   unsigned int base;
   DIRA |= i2c_pins; // i2c pins as outputs
   base = delay + CNT; // reference time for sync delay timing of scl edges
                       // (ref propeller manual, pg 325 for description of technique) 
   OUTA &= scl_pinb; // scl low
   // send dat, msb to lsb
   for(i = 0; i < 8; ++i) {
      // change sda while scl low
      if (dat & 0x80) {
	     OUTA |= sda_pin;
	  } else {
	     OUTA &= sda_pinb;
	  }
	  WAITCNT(base, delay);
      OUTA |= scl_pin; // scl high
	  dat <<= 1;
	  WAITCNT(base, delay);
	  OUTA &= scl_pinb; // scl low
   }
   // now receive ack: sda to input
   DIRA &= sda_pinb;
   WAITCNT(base, delay);
   OUTA |= scl_pin; // scl high
   // read sda while scl high
   ack = ((INA & sda_pin) == sda_pin);
   WAITCNT(base, delay);
   OUTA &= scl_pinb; // scl low
   WAITCNT(base, delay);
   // leave sda low
   DIRA |= sda_pin;
   OUTA &= sda_pinb;
   WAITCNT(base, delay);

   return ack;
}

/** read 1 byte from the i2c bus */
static char i2cRead(char ack)
{
   // Read in i2c data, Data byte is output MSB first, SDA data line is valid
   // only while the SCL line is HIGH
   int dat = 0;
   int i;
   const unsigned int delay = i2c_delay; // local copy of i2c_delay for use in WAITCNT
   unsigned int base;
   // set scl to output, sda to input
   DIRA |= scl_pin;
   DIRA &= sda_pinb;
   base = delay + CNT;
   OUTA &= scl_pinb; // scl low
 
   // read in 1 byte, msb to lsb
   for(i = 0; i < 8; ++i) {
     WAITCNT(base, delay);
     OUTA |= scl_pin; // scl high
	 dat <<= 1;
	 dat |= ((INA & sda_pin) == sda_pin); // read sda while scl high
	 WAITCNT(base, delay);
     OUTA &= scl_pinb; // scl low
   }
   // send ack/nack (while scl low)
   DIRA |= sda_pin;
   if (ack)
     OUTA |= sda_pin;
   else
     OUTA &= sda_pinb;
   WAITCNT(base, delay);
   OUTA |= scl_pin; // scl high
   WAITCNT(base, delay);
   OUTA &= scl_pinb; // scl low
   WAITCNT(base, delay);
   return dat;
}

/** initialize the nunchuck. Needed before every read operation
 * @return 0 if nunchuck OK, non-zero otherwise
 */
static int i2cNunchuckWriteDev()
{
   int rc = 0;

   i2cStart();
   rc = 0;
   rc = (rc << 1) | i2cWrite(NUNCHUCK_ADDR);
   rc = (rc << 1) | i2cWrite(0x40);
   rc = (rc << 1) | i2cWrite(0x00);
   i2cStop();
   wait(nunchuck_cmd_delay); // required delay for nunchuck to process cmd

   return rc;
}

/** initialize a Wii nunchuck on the i2c bus
  * @param pinSDA	 i2c data pin (green wire)
  * @param pinSCL    i2c clock pin (yellow wire)
  * @return 0 if nunchuck successfully initialized, non zero otherwise
  */
int Nunchuck_init(int pinSDA, int pinSCL)
{
   // remember the i2c pins we'll be using
   scl_pin = 1 << pinSCL;
   sda_pin = 1 << pinSDA;
   // setup the inverse bitmasks for the pins (to simplify setting pin low)
   scl_pinb = ~scl_pin;
   sda_pinb = ~sda_pin;
   i2c_pins = scl_pin | sda_pin;
   // to begin with, i2c pins are inputs
   DIRA |= i2c_pins;
   OUTA |= i2c_pins; // initialize both pins high (ie same as i2c pullups)

   nunchuck_cmd_delay  = (CLKFREQ / 100000)*30; // 300uSecs between cmds is needed
   i2c_delay = CLKFREQ / (I2C_FREQ * 2); // calculate edge to edge delay for scl
   
   return i2cNunchuckWriteDev();
}

/** read in the current Wii nunchuck sensor data
  * and store it into the datastructure "nd"
  * @param nd Datastructure into which to store Nunchuck results
  * @return 0 if successful, non zero otherwise
  */
int Nunchuck_read(NunchuckDatT *nd)
{
   unsigned char rawData[6];
   int rc = i2cNunchuckWriteDev(); // initialize nunchuck
   // request data from nunchuck
   i2cStart();
   i2cWrite(NUNCHUCK_ADDR);
   i2cWrite(0);
   i2cStop();
   wait(nunchuck_cmd_delay); // required delay for nunchuck to process cmd
   // read 6 bytes of data from nunchuck
   i2cStart();
   i2cWrite(NUNCHUCK_ADDR|1);
   // nunchuck data is encoded -> decode each byte as we receive it 
   rawData[0] = (i2cRead(0) ^ 0x17) + 0x17; // read with ACK for first 5 bytes
   rawData[1] = (i2cRead(0) ^ 0x17) + 0x17;
   rawData[2] = (i2cRead(0) ^ 0x17) + 0x17;
   rawData[3] = (i2cRead(0) ^ 0x17) + 0x17;
   rawData[4] = (i2cRead(0) ^ 0x17) + 0x17;
   rawData[5] = (i2cRead(1) ^ 0x17) + 0x17; // last byte: read with NACK
   i2cStop();
   // Now unpack the raw data into more readable structure
   nd->joyX = rawData[0];
   nd->joyY = rawData[1];
   nd->buttonZ = ((rawData[5] & 0x1) == 0x0);
   nd->buttonC = ((rawData[5] & 0x2) == 0x0);
   // accel data is 10 bits: 8 bits in the appr rawData slot, + 2 bits in rawData[5]
   nd->accelX = (rawData[2] << 2) + ((rawData[5] & 0b00001100) >> 2); 
   nd->accelY = (rawData[3] << 2) + ((rawData[5] & 0b00110000) >> 4);
   nd->accelZ = (rawData[4] << 2) + ((rawData[5] & 0b11000000) >> 6);
   return rc;
}

/*
+------------------------------------------------------------------------------------------------------------------------------+
                                                   TERMS OF USE: MIT License                                                                                                              
+------------------------------------------------------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation     
files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    
modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:                                                                   
                                                                                                                              
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
                                                                                                                              
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         
+------------------------------------------------------------------------------------------------------------------------------+
*/
